float f_0_0(void);
double d_0_0(void);
float f_m0_0(void);
double d_m0_0(void);
float f_0_17(void);
double d_0_17(void);
float f_m0_17(void);
double d_m0_17(void);
float f_1_0(void);
double d_1_0(void);
float f_m1_0(void);
double d_m1_0(void);
float f_1_7(void);
double d_1_7(void);
float f_m1_7(void);
double d_m1_7(void);
float f_42(void);
double d_42(void);
float f_m42(void);
double d_m42(void);
float f_nan(void);
double d_nan(void);
float f_nan_x(void);
double d_nan_x(void);
float f_inf(void);
double d_inf(void);
float f_m_inf(void);
double d_m_inf(void);
